# MIT License

# Copyright (c) 2020 yaq

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Our new and improved APT protocol unpacker, which we'll include
# here until (if?) our merge request gets accepted.

__all__ = ["Unpacker"]

import asyncio
from collections import namedtuple
import io
import struct
import warnings

from .parsing import id_to_func


class Unpacker:
    """
    Create an Unpacker to decode a byte stream into Thorlabs APT protocol messages.

    The ``file_like`` parameter should be an object which data can be sourced from.
    It should support the ``read()`` method.

    The ``on_error`` parameter selects the action to take if invalid data is detected.
    If set to ``"continue"`` (the default), bytes will be discarded if the byte sequence
    does not appear to be a valid message.
    If set to ``"warn"``, the behaviour is identical, but a warning message will be emitted.
    To instead immediately abort the stream decoding and raise a ``RuntimeError``, set to
    ``"raise"``.

    :param file_like: A file-like object which data can be `read()` from.
    :param on_error: Action to take if invalid data is detected.
    """
    def __init__(self, file_like=None, on_error="continue"):
        if file_like is None:
            self._file = io.BytesIO()
        else:
            self._file = file_like
        self.buf = b""
        self.on_error = on_error

    def __iter__(self):
        return self

    def _decoding_error(self, message="Error decoding message from buffer."):
        """
        Take appropriate action if parsing of data stream fails.

        :param message: Warning or error message string.
        """
        if self.on_error == "raise":
            raise RuntimeError(message)
        if self.on_error == "warn":
            warnings.warn(message)
        # Discard first byte of buffer, it might decode better now...
        self.buf = self.buf[1:]

    def __next__(self):
        # Basic message packet is 6 bytes, try to fill buffer to at least that size
        if len(self.buf) < 6:
            self.buf += self._file.read(6 - len(self.buf))
        # Hopefully enough data in buffer now to try to decode a message
        while len(self.buf) >= 6:
            # Look at first two bytes and ensure they look like a message ID we recognise
            msgid, length = struct.unpack_from("<HH", self.buf)
            if not msgid in id_to_func:
                self._decoding_error(f"Invalid message with id={msgid:#x}")
                continue
            # Looks like a message, now check the source and destination locations
            long_form = self.buf[4] & 0x80  # Check MSB of byte 4 for "long form" flag
            dest = self.buf[4] & ~0x80      # Destination is remaining lower bits
            source = self.buf[5]
            # Destination should be the Host, source should be a recognised controller ID
            if not ((dest in (0x00, 0x01)) and (source in (0x00, 0x11, 0x21, 0x22, 0x23, 0x24, 0x25,
                                                           0x26, 0x27, 0x28, 0x29, 0x2A, 0x50))):
                self._decoding_error("Invalid source or destination for message with id="
                                    f"{msgid:#x}, src={source:#04x}, dest={dest:#04x}")
                continue
            # Message ID, source and dest seem legit, now check long form length
            if long_form:
                # A bad or malicious packet could make us try to read up to 65 kB...
                # Documentation says "currently no datapacket exceeds 255 bytes in length"
                if length > 255:
                    self._decoding_error(f"Invalid length={length} for message with id={msgid:#06x},"
                                         f" src={source:#04x}, dest={dest:#04x}")
                    continue
            else:
                # Length field is actually two parameters in short form messages
                length = 0
            # Either short form message, or long form message of reasonable size
            # Looks good! Break from loop and proceed
            break                    
        # If we got here, either the buffer was/shrank too small,
        # or we have the start of something that looks like a valid message
        if len(self.buf) < 6:
            # Not enough data to form a message packet
            raise StopIteration
        # Buffer contains enough for a short message, but maybe not a long form one
        if len(self.buf) < length + 6:
            # Not enough data in buffer to decode long form message, attempt to read some more data
            self.buf += self._file.read(length - len(self.buf) + 6)
            if len(self.buf) < length + 6:
                # Still didn't receive enough data to decode message
                raise StopIteration
        # Have enough data in buffer to decode the full message
        data = self.buf[:length + 6]
        # Can now remove the message data from the buffer
        self.buf = self.buf[length + 6:]
        # Decode the message contents
        dict_ = id_to_func[msgid](data)
        return namedtuple(dict_["msg"], dict_.keys())(**dict_)

    def __aiter__(self):
        return self

    async def __anext__(self):
        while True:
            try:
                return next(self)
            except StopIteration:
                await asyncio.sleep(0.001)

    def feed(self, data: bytes):
        """
        Add byte data to the input stream.

        The input stream must support random access, if it does not, must be fed externally
        (e.g. serial port data).

        :param data: Byte array containing data to add.
        """
        pos = self._file.tell()
        self._file.seek(0, 2)
        self._file.write(data)
        self._file.seek(pos)



# Copyright 2021 Patrick C. Tapping
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ["BSC201", "BSC201_DRV250"]

from .. import protocol as apt
from .aptdevice_motor import APTDevice_BayUnit
from ..enums import EndPoint


class BSC201(APTDevice_BayUnit):
    """
    A class for ThorLabs APT device model BSC201.

    It is based off :class:`BBD_BSC` configured with a single channel and the BSC the serial number
    preset in the initialisation method.

    Note that it seems the BSC201 will send automatic status updates, but does not accept the 
    corresponding acknowledgement message, and so will stop responding after ~5 seconds if they are
    enabled. As a workaround, the status of the device will be polled.

    Additionally, it seems that the initial movement and homing velocities can be effectively zero,
    making it seem like the device is not working, though it is actually just moving extremely
    slowly.
    It's a good idea to ensure sensible values are set for :meth:`set_velocity_params`, :meth:`set_jog_params`
    and :meth:`set_home_params` during initialisation.

    :param serial_port: Serial port device the device is connected to.
    :param vid: Numerical USB vendor ID to match.
    :param pid: Numerical USB product ID to match.
    :param manufacturer: Regular expression to match to a device manufacturer string.
    :param product: Regular expression to match to a device product string.
    :param serial_number: Regular expression to match to a device serial number.
    :param home: Perform a homing operation on initialisation.
    :param invert_direction_logic: Invert the meaning of "forward" and "reverse" directions.
    :param swap_limit_switches: Swap "forward" and "reverse" limit switch values.
    """
    def __init__(self, serial_port=None, vid=None, pid=None, manufacturer=None, product=None, serial_number="40", location=None, home=True, invert_direction_logic=False, swap_limit_switches=True):

        super().__init__(serial_port=serial_port, vid=vid, pid=pid, manufacturer=manufacturer, product=product, serial_number=serial_number, location=location, x=1, home=home, invert_direction_logic=invert_direction_logic, swap_limit_switches=swap_limit_switches, status_updates="polled")


class BSC201_DRV250(BSC201):
    """
    A class for ThorLabs APT device model BSC201 with DRV250 stepper-motor-driven actuator, sold as
    a package as the `LNR502 and LNR502E <https://www.thorlabs.com/newgrouppage9.cfm?objectgroup_id=2297&pn=LNR502E/M>`__.

    It is based off :class:`BSC201`, but with sensible default movement parameters configured for the actuator.

    For the DRV250, there are 409600 microsteps/mm, 21987328 microsteps/mm/s, and 4506 microsteps/mm/s/s.

    :param serial_port: Serial port device the device is connected to.
    :param vid: Numerical USB vendor ID to match.
    :param pid: Numerical USB product ID to match.
    :param manufacturer: Regular expression to match to a device manufacturer string.
    :param product: Regular expression to match to a device product string.
    :param serial_number: Regular expression to match to a device serial number.
    :param home: Perform a homing operation on initialisation.
    :param invert_direction_logic: Invert the meaning of "forward" and "reverse" directions.
    :param swap_limit_switches: Swap "forward" and "reverse" limit switch values.
    """
    def __init__(self, serial_port=None, vid=None, pid=None, manufacturer=None, product=None, serial_number="40", location=None, home=True, invert_direction_logic=False, swap_limit_switches=True):

        super().__init__(serial_port=serial_port, vid=vid, pid=pid, manufacturer=manufacturer, product=product, serial_number=serial_number, location=location, home=home, invert_direction_logic=invert_direction_logic, swap_limit_switches=swap_limit_switches)

        # Initial velocity parameters are effectively zero on startup, set something more sensible
        # Homing is initiated 1.0s after init, so hopefully these will take effect before then...
        for bay_i, _ in enumerate(self.bays):
            for channel_i, _ in enumerate(self.channels):
                self.set_velocity_params(acceleration=4506, max_velocity=21987328, bay=bay_i, channel=channel_i)
                self.set_jog_params(size=409600, acceleration=4506, max_velocity=21987328, bay=bay_i, channel=channel_i)
                self.set_home_params(velocity=21987328, offset_distance=20480, bay=bay_i, channel=channel_i)
       



# Copyright 2021 Patrick C. Tapping
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Python interface to ThorLabs devices which use the APT protocol.

Note that everything important from the submodules is imported directly here to the main package.
This means you may use code like ``from thorlabs_apt_device import APTDevice`` (instead of
using the full ``from thorlabs_apt_device.devices.aptdevice import APTDevice`` or similar).
"""

__version__ = "0.3.2"

from .devices import *
from .enums import *
from .utils import *



#!/usr/bin/env python3
import os
import sys
import shutil
import setuptools

# Workaround issue in pip with "pip install -e --user ."
import site
site.ENABLE_USER_SITE = True

with open("README.rst", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name="thorlabs_apt_device",
    version="0.3.2",
    author="Patrick Tapping",
    author_email="mail@patricktapping.com",
    description="Interface to ThorLabs devices which communicate using the APT protocol.",
    long_description=long_description,
    url="https://gitlab.com/ptapping/thorlabs-apt-device",
    project_urls={
        "Documentation": "https://thorlabs-apt-device.readthedocs.io/",
        "Source": "https://gitlab.com/ptapping/thorlabs-apt-device",
        "Tracker": "https://gitlab.com/ptapping/thorlabs-apt-device/-/issues",
    },
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=[
        "pyserial",
    ],
)
